#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed May 27 17:21:12 2020

@author: yves

    This file is part of Saqqarah.

    Saqqarah is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Saqqarah is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Saqqarah.  If not, see <https://www.gnu.org/licenses/>
"""

import numpy as np
import itertools
from time import time, sleep

class Pyramid():
    """ 
    Pyramid with size and number of boxes
    Method get_puzzle give a puzzle with solution
    
    Not clean: rewrtie with getaatr or properties
    """
    
    # stolen in the net (choose)
    # 
    def nCr(self, n, k):
        """
        A fast way to calculate binomial coefficients by Andrew Dalke (contrib).
        """
        if 0 <= k <= n:
            ntok = 1
            ktok = 1
            for t in range(1, min(k, n - k) + 1):
                ntok *= n
                ktok *= t
                n -= 1
            return ntok // ktok
        else:
            return 0
        
    def log(self, msg, *, end='\n'):
        if not self.ui:
            print(msg, end=end)
        else:
            self.ui.information(msg)

    def __init__(self, size, ui=False):
        if size < 4:
            raise ValueError("Pyramid Size at least 4")
        # size
        self.n = size
        # run with ui
        self.ui = ui
        self.tries = 0
        # n lines of the matrix
        self.n_lines = self.n*(self.n+1)//2
        self.n_comb = self.nCr(self.n_lines, self.n)
        # calculation matrix from base line to all values
        self.mat = np.empty((self.n_lines,self.n), dtype='int')
        
        # used only to statistical fonctions 
        self.determinants = None

        m_line = 0
        for part in range(self.n,0,-1):
            n_comb = self.n - part
            for line in range(part):
                for case in range(self.n):
                    self.mat[m_line, case] = self.nCr( n_comb, case - line)
                m_line += 1
                
        self.solution = []
        self.puzzle = {}
        
    def get_size(self):
        return self.n
    
    def get_nbr_cases(self):
        return self.n_lines
    
    def get_puzzle(self, difficulty = 2, min=1, max = 9, *, seed = None, exclude_zero = False):
        """
        return a dict { number_of_case : value }
        you can use it to set a puzzle at hand
        """
        if difficulty < 1 or difficulty > 8:
            raise ValueError("difficulty between 1 and 8")
        elif difficulty >  3 and self.n == 4:
            raise ValueError("size too low for difficulty")

        # self.random_cases are cases to fill for the puzzle
        # get nanoseconds of time as reusable seed as fallback
        self.seed = seed or int(str(round(time()*1_000_000))[-9:])
        np.random.seed(seed)
        self.random_cases = np.array(self.__choose_random__(difficulty))
        self.random_cases.transpose()

        # choose values for base line
        if exclude_zero:
            if max > 0:
                # random on [min, max-1] & replace 0 by max
                self.base_solution = np.random.randint(min, max, size=(self.n,1))
                self.base_solution[self.base_solution == 0] = max
            else: # max == 0
                self.base_solution = np.random.randint(min+1, max+1, size=(self.n,1))
                self.base_solution[self.base_solution == 0] = min
        else:
            self.base_solution = np.random.randint(min, max+1, size=(self.n,1))
        
        # get values for all cases
        self.solution = np.matmul(self.mat, self.base_solution)

        # get index and values for selected cases
        # the visible part of the puzzle
        self.puzzle = { c: int(self.solution[c,0]) for c in self.random_cases }
        
        return self.puzzle

    def __choose_random__(self, difficulty = 2):
        """
        choose at random cases to fill for given difficulty
        """
        lines = list(range(self.n_lines))
        np.random.shuffle(lines)
        iter_mat = itertools.combinations(lines, self.n)
        
        self.cancel = False
        det = 0
        for self.tries,l in enumerate(iter_mat):
            det =  round(abs(np.linalg.det(self.mat[l,])))
            if det == difficulty:
                tries='try' if self.tries==0 else 'tries'
                self.log(f'Find a puzzle for difficulty {difficulty}, total {self.tries+1} {tries}')
                return tuple(sorted(l))
            if self.tries%100 == 0:
                self.log(f"Looking for puzzle... {self.tries}", end='\r')
            # break by UI cancel.
            if self.cancel:
                return tuple()
            #sleep(0.000_001)
        else:
            raise ValueError("This difficulty was not found with that size")
    
    #####
    # statistical only calculations: all determinants
    # for all choices of cases
    # only det !=0 give a possible puzzle 
    # abs(determinant) is used as a difficulty level 
    #####     
    def __get_determinants__(self):
        """
        list all determinants in self.determinants
        can be long, 9min on my old core i7 for n=8
        """
        lines = list(range(self.n_lines))
        # random.shuffle(lines)
        sel_mat = itertools.combinations(lines, self.n)
        self.determinants =  {}

        for l in sel_mat:
            #l = sorted(l)
            det =  abs(round(np.linalg.det(self.mat[l,])))
            if det in self.determinants:
                self.determinants[det].append(l)
            else:
               self.determinants[det] = [l]
               
    def __nbr_cas__(self):
        return self.nCr(self.n_lines, self.n_lines - self.n)

    def __bilans__(self):
        # can be long
        if not self.determinants:
            self.__get_determinants__()
            
        if 0 in self.determinants:
            n_zeros = len(self.determinants[0])
        else:
            n_zeros = 0
        self.log(f"{self.__nbr_cas__()} choix possibles dont {n_zeros} impossibles.")
        max_det = max(self.determinants.keys())
        self.log(f"Déterminant maximal : {max_det}")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun  3 17:59:23 2020

@author: yves


    This file is part of Saqqarah.

    Saqqarah is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Saqqarah is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Saqqarah.  If not, see <https://www.gnu.org/licenses/>
"""

import sys
from PySide2 import QtCore, QtGui, QtWidgets
from PySide2.QtUiTools import QUiLoader
from PySide2.QtCore import SIGNAL

from saqqarah import import SaqqarahUI


loader = QUiLoader()


def print_dict(obj):
    for k,v in obj.__dict__.items():
        print(f"{k} : {v}")

def mainwindow_setup(w):
    w.setWindowTitle("MainWindow Title")

app = QtWidgets.QApplication.instance()
if app is None: 
    app = QtWidgets.QApplication(sys.argv)

window = loader.load("saqqarah-qt.ui", None)
uicontrol = SaqqarahUI(window) 

mainwindow_setup(window)
window.show()

window.quitButton.connect(SIGNAL ('clicked()'), app.quit)
#print_dict(window)
sys.exit(app.exec_())




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jun 12 23:27:46 2020

@author: yves


    This file is part of Saqqarah.

    Saqqarah is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Saqqarah is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Saqqarah.  If not, see <https://www.gnu.org/licenses/>
"""
from . import ui
try:
    import importlib.resources as pkg_resources
except ImportError:
    # Try backported to PY<37 `importlib_resources`.
    import importlib_resources as pkg_resources
    
from PySide2.QtUiTools import QUiLoader
    
class PreferencesUI:
    resolutions = [ 50, 75, 100, 150, 200, 250, 300, 600 ]
    
    def __init__(self, ui_instance):
        self.ui = ui_instance
        
        loader = QUiLoader()
        with pkg_resources.path(ui, 'preferences.ui') as ui_file:
            self.pref_ui = loader.load(str(ui_file), self.ui.window)
        
        self.pref_ui.show()
        self.resolution_init()
        self.timestamp_init()
        self.linewidth_init()
        self.basename_init()
        self.buttonBox = self.pref_ui.buttonBox
            
    def resolution_init(self):
        
        if self.ui.param.resolution not in self.resolutions:
            self.resolutions.insert(0, self.ui.param.resolution)
            self.resolutions.sort()
        
        actual_index = self.resolutions.index(self.ui.param.resolution)
        
        resolutionCombo = self.pref_ui.resolutionCombo
                
        resolutionCombo.addItems(str(r) for r in self.resolutions)
        resolutionCombo.setCurrentIndex(actual_index)
        
    def timestamp_init(self):
        names_choice= [ 'Add timestamp in file name',
                       'Erase old file when save' ]
        actual_index = 0 if self.ui.param.use_timestamp else 1

        stampCombo = self.pref_ui.timestampCombo
        stampCombo.addItems(names_choice)
        stampCombo.setCurrentIndex(actual_index)
        
    def linewidth_init(self):
        lwbox = self.pref_ui.linewidthBox
        lwbox.setValue(self.ui.param.line_width)
        
    def basename_init(self):
        self.pref_ui.basenameEdit.setPlaceholderText(self.ui.param.basename)
        


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat May 30 00:01:30 2020

@author: yves

    This file is part of Saqqarah.

    Saqqarah is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Saqqarah is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Saqqarah.  If not, see <https://www.gnu.org/licenses/>
"""

# expose directly the classes in the module
from .version import version, codename
from .tools import *
from .ParametersClass import Parameters, PrinterCoordinates
from .PyramidClass import Pyramid
from .PyramidPrinterClass import PyramidPrinter
from .PyramidPrinterTikzClass import PyramidPrinterTikz
from .PyramidPrinterImageClass import PyramidPrinterImage
from .PreferencesUIClass import PreferencesUI
from .SaqqarahUIClass import SaqqarahUI




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri May 29 23:37:07 2020

@author: yves

    This file is part of Saqqarah.

    Saqqarah is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Saqqarah is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Saqqarah.  If not, see <https://www.gnu.org/licenses/>
"""

from . import PyramidPrinter
from . import PrinterCoordinates
from . import filename_prepare, ensure_attr

from time import strftime
from pathlib import Path

import os, sys

class PyramidPrinterTikz(PyramidPrinter):
    """ 
    implements printing in tikz
    """
    def __init__(self, pyramid, param):
        super().__init__(pyramid, param)
        param.default_extension = 'tex'
        self.unit = 'cm'

    def print_pyramid(self, param):
        for i in range(param.size):
            y = param.size - i 
            self.__print_line__(y, i+1, param)

    def __add_line_out__(self, line, param):
        param.printer.put(line, param)

    def __print_line__(self, y, no, param):
        for x in range(param.size-no, param.size+no,2):
            self.__print_box__(x,y, param)
            self.__print_node__(x,y, param)
            
    def __print_box__(self, x, y, param):
        # Python is cool
        c1, c2, c3, c4 = (PrinterCoordinates(param) for c in range(4))
        c1.xy = x, y
        c2.xy = x + param.boxw, y
        c3.xy = x + param.boxw, y - param.boxh
        c4.xy = x, y - param.boxh
        self.__add_line_out__(fr"\draw [line width={param.line_width}pt] ({c1})--({c2})--({c3})--({c4})--cycle;", param)

    def __print_node__(self, x, y, param):
        if not param.values:
            return
        value = param.values.pop()
        if value == "_":
            return

        c = PrinterCoordinates(param)
        c.xy = x + param.boxw/2, y - param.boxh/2

        # puzzle
        if type(value) is str:
            color="[text=red]"
        else:
            color=""
        self.__add_line_out__(f"\\draw ({c}) node{color} \u007b ${value}$ \u007d ;", param)

    class PrinterCM(PyramidPrinter.PrinterCM):
        def  __init__(self, printer):
            # if filename does not open to a file clone sys.stdout
            self.param = printer.param
            self.out = { stream: "" for stream in self.param.streams }
 
        def __enter__(self):
           # dict of out strings for each pyramid
           return self
            
        def __exit__(self, *args):
            filename = self.param.get_filename()
            try:
                self.file = open(filename, "w")
            except Exception as e:
                # PermissionError or FileNotFoundError
                print(f"Error {e}: {filename}")
                filename=None
                self.file = os.fdopen(os.dup(sys.stdout.fileno()),"w")

            with self.file as f:
                for stream in self.param.streams:
                    f.write(self.out[stream])
                    f.write("\n")
            if filename:
                self.param.log(f"Tikz code written in {filename}")
            else:
                self.param.log(f"% Tikz code written in standard output.")
                    
        def put(self, line, param):
            self.out[param.stream] += line + '\n'

    class PyramidCM(PyramidPrinter.PyramidCM):
        def __init__(self, printer):
            self.param = printer.param
        
        def __enter__(self):
            stream = self.param.stream

            #add eventually a separator before each pyramid but first
            index = self.param.index
            ensure_attr(self.param, 'latex_separator', '')
            if index > 0:
                self.param.printer.put(self.param.latex_separator, self.param)

            # begin tikz image
            self.param.printer.put(f"%% {stream}", self.param)
            self.param.printer.put(r"\begin{tikzpicture}", self.param)

        def __exit__(self, *args):
            # end tikz image
            self.param.printer.put(r"\end{tikzpicture}", self.param)
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jun  4 22:36:13 2020

@author: yves
"""

import os
from pathlib import Path
   
def filename_prepare(name, extensions, defaut_extension):
    """ 
    return basename, extension tuple
    keeping extension given if it's in extensions list
    add default_extension if not
    """
    
    if type(defaut_extension) is str:
        defaut_extension = defaut_extension.lower()
        
    if not defaut_extension in extensions:
        extensions.insert(0, defaut_extension)
    
    # replace spaces by _ and then split by '.'
    name = '_'.join(name.split())
    
    # check if name contains directory
    # os.path.dirname returns '' is there is no dirname
    # None if no dirname
    dirname = os.path.dirname(name) or None
    
    if dirname and not Path(dirname).is_dir():
            dirname = None
    
    pname = Path(name)
    # in pathlib, suffix contains the '.'
    ext = pname.suffix[1:].lower()
    basename = pname.stem
    
    # extension is in extensions_list ?
    if not ext in [ extension.lower() for extension in extensions]:
        # basenme is full name
        # extension is default_extension
        basename = pname.name
        ext = defaut_extension
        
    return basename, ext, dirname

def ensure_attr(obj, key, default):
    """
    put a default attribute if it does not exists or is None
    """
    if hasattr(obj, key):
        if getattr(obj,key) is None:
            setattr(obj,key, default)
    else:
        setattr(obj,key, default)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri May 29 23:33:40 2020

@author: yves

    This file is part of Saqqarah.

    Saqqarah is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Saqqarah is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Saqqarah.  If not, see <https://www.gnu.org/licenses/>
"""

from . import Parameters, ensure_attr
        
class PyramidPrinter:
    """
        base class for printing pyramid
        
        mask in subclass:
            - PrinterCM: contextmanager for all printing
            - PyramidCM: context manager for one pyramid printing
    """
    def __init__(self, pyramid, param):
        self.pyramid = pyramid
        self.param = param

    class PrinterCM:
        def __init__(self, printer):
            """
            Initialisation code for the printer itself
            **kwargs are passer from self.printer
            Implement in subclass
            """
            raise NotImplementedError("Don't use directly this class !")

        def __enter__(self):
            """
            Enter code for the printer
            Implement in subclass
            """
            raise NotImplementedError("Don't use directly this class !")

        def __exit__(self, *args):
            """
            finalization code for the printer
            Implement in subclass
            """

            raise NotImplementedError("Don't use directly this class !")

    class PyramidCM:
        def __init__(self, printer):
            """
            Initialisation code for the pyramid printing
            **kwargs are passer from self.printer
            Implement in subclass
            """
            raise NotImplementedError("Don't use directly this class !")

        def __enter__(self):
            """
            Enter code for the pyramid printing
            Implement in subclass
            """
            raise NotImplementedError("Don't use directly this class !")

        def __exit__(self, *args):
            """
            Enter code for the printer printing
            Implement in subclass
            """
            raise NotImplementedError("Don't use directly this class !")
            

    def print(self):
        # check unit is ok
        if hasattr(self, 'unit'):
            self.param.unit = self.unit

        # TODO: UGLY ! CLEAN THAT
        self.param.pyramid_puzzle = self.pyramid.puzzle
        self.param.pyramid_solution  = self.pyramid.solution
        self.param.log = self.pyramid.log

        # puzzle and solution are boolean to know
        # what must be printed
        # TODO clean that (print_puzzle e.g.)        
        puzzle = self.param.puzzle if hasattr(self.param, 'puzzle') else True
        solution = self.param.solution if hasattr(self.param, 'solution') else True
        
        if puzzle and not self.param.pyramid_puzzle:
            puzzle = False
        if solution and self.param.pyramid_solution == []:
            solution = False

        self.param.streams = ['NoPuzzle'] if not puzzle and not solution else []
        self.param.streams += ['Puzzle'] if puzzle else []
        self.param.streams += [ 'Solution' ] if solution else []
        
        with self.PrinterCM(self) as printer:
            for index, stream in enumerate(self.param.streams):
                self.param.values = self.__set_values__(stream)
                self.param.index = index
                self.param.stream = stream
                self.param.printer = printer

                with self.PyramidCM(self) as pyramid_printer:
                    self.param.pyramid_printer = pyramid_printer
                    self.print_pyramid(self.param)

    def __set_values__(self, stream):
        """
        prepare list of values to print in cases for output
        """
        if stream == 'Puzzle':
            values = []
            for k in range(self.param.n_values()):
                values.append(
                        self.param.pyramid_puzzle[k] 
                        if k in self.param.pyramid_puzzle 
                        else '_'
                        )
        elif stream == 'Solution':  
            values = [ int(n) for n in self.param.pyramid_solution ]
            # change type if case is in puzzle
            # str will be type of puzzle values
            # used to change color in solution
            for k in self.param.pyramid_puzzle:
                values[k] = str(values[k])
        else:
            values = ['_'] * self.param.n_values()
        return values
                        
    def print_pyramid(self):
        """
        Implement in the subclass
        self.values as values to print, or '_'
        self.size as size of the pyramid
        self.current_stream is 'puzzle' or 'solution'
        """
        raise NotImplementedError("Don't use directly this class !")
           
            
        


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun  3 22:34:32 2020

@author: yves


    This file is part of Saqqarah.

    Saqqarah is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Saqqarah is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Saqqarah.  If not, see <https://www.gnu.org/licenses/>
"""

from . import Pyramid, PyramidPrinterImage, PyramidPrinterTikz

from PySide2 import QtCore, QtGui, QtWidgets
from PySide2.QtCore import QTimeLine, QThread
from PySide2.QtGui import QRegExpValidator
from PySide2.QtWidgets import QFileDialog, QAction

from PIL.ImageQt import toqpixmap

from . import PreferencesUI
from . import Parameters

from PySide2.QtUiTools import QUiLoader

from . import version, codename

from . import ui
try:
    import importlib.resources as pkg_resources
except ImportError:
    # Try backported to PY<37 `importlib_resources`.
    import importlib_resources as pkg_resources
    
class LookingForPuzzleThread(QThread):
    def __init__(self, ui):
        QThread.__init__(self)
        self.param = ui.param
        self.pyramid = ui.pyramid

    def __del__(self):
        self.wait()

    # TODO : pyramid uses parameters() too
    def run(self):
        self.pyramid.get_puzzle(difficulty=self.param.difficulty,
                                min = self.param.value_min,   
                                max = self.param.value_max,
                                seed = self.param.seed,
                                exclude_zero = self.param.exclude_zero)

class SaqqarahUI:
    def __init__(self, window):
        # defaut values are inside
        self.param = Parameters()

        self.images = {}

        self.generating = False

        self.init_ui(window)
        # 1000 seconds
        self.timeline = QTimeLine(2_000, self.window)
        self.init_timeline()
        self.combo_init()
        self.seed_init()
        self.zero_init()
        self.update_pyramid()
        
    def zero_init(self):
        zeroCheck = self.window.zeroCheckbox
        self.exlude_zero = zeroCheck.isChecked()
        zeroCheck.stateChanged.connect(self.zero_changed)

    def zero_changed(self):
        self.param.exclude_zero = self.window.eroCheckbox.isChecked()

    def seed_init(self):
        seed = self.window.seedEdit
        seed.setValidator(QRegExpValidator("[1-9][0-9]*"))
        seed.editingFinished.connect(self.seed_changed)

    def seed_changed(self):
        value = self.window.seedEdit.text()
        self.param.seed = int(value)

    def combo_init(self):
        outCombo = self.window.outputTypeCombo
        outCombo.addItem('Images', userData=PyramidPrinterImage)
        outCombo.addItem('Tikz code', userData=PyramidPrinterTikz)

    def update_pyramid(self):        
        self.pyramid = Pyramid(self.param.size, ui=self)
        self.imagePrinter = PyramidPrinterImage(self.pyramid, self.param, memory_images=True)
        self.clean_images()
           
    def init_ui(self,window):
        self.window = window
        self.window.quitButton.clicked.connect(QtWidgets.QApplication.instance().quit)
        self.spinbox_init(window.sizeBox, 
                          value=self.param.size, 
                          min=self.param.size_min, 
                          max=self.param.size_max, 
                          on_change=self.size_changed )
        self.check_size_difficulty()

        self.spinbox_init(window.difficultyBox, 
                          value=self.param.difficulty, 
                          min=self.param.difficulty_min, 
                          max=self.param.difficulty_max,
                          on_change=self.difficulty_changed )
        self.spinbox_init(window.minBox, 
                          value=self.param.value_min, 
                          min=-1000, 
                          max=self.param.value_max-1,
                          on_change=self.min_changed )
        self.spinbox_init(window.maxBox, 
                          value=self.param.value_max, 
                          min=self.param.value_min+1, 
                          max=1000,
                          on_change=self.max_changed )
        self.window.puzzleButton.clicked.connect(self.generate_puzzle)
        self.window.directoryButton.clicked.connect(self.choose_directory)
        self.window.saveButton.clicked.connect(self.save_puzzle)
        self.progressbar_init()
        self.menus_init()
        
    def menus_init(self):
        entryExit = self.window.actionExit
        entryExit.triggered.connect(QtWidgets.QApplication.instance().quit)
        self.window.actionPreferences.triggered.connect(self.preferences_run)
        self.window.actionAbout.triggered.connect(self.about_run)
        
    def about_run(self):
        loader = QUiLoader()
        
        with pkg_resources.path(ui, 'about.ui') as ui_file:
            about_ui = loader.load(str(ui_file), self.window)
            
        about_ui.versionLabel.setText(f"Version {version()} ({codename()})")
        about_ui.show()
        
    def preferences_run(self):
        self.pref_ui = PreferencesUI(self).pref_ui
        self.pref_ui.accepted.connect(self.validate_pref)
  
    def init_timeline(self):
        # 0.1 sec frame
        self.timeline.setUpdateInterval(10)
        self.timeline.setFrameRange(0, 100)
        self.timeline.frameChanged[int].connect(self.update_progressbar)
        #2**12 * 10 seconds is 11 hours
        self.timeline.setLoopCount(2**15)
        self.timeline.setCurveShape(self.timeline.LinearCurve)

    def update_progressbar(self, frame):
        # animation to make sure ui is not locked
        X = frame*2 - 100
        self.window.progressBar.setInvertedAppearance(X >= 0)
        X = 100 - abs(X)
        self.window.progressBar.setValue(X)
     
    def progressbar_init(self):
        self.window.progressBar.setRange(0,100)
        self.window.progressBar.setInvertedAppearance(False)
        self.window.progressBar.reset()

    def progressbar_stop(self):
        self.timeline.stop()
        self.window.progressBar.setInvertedAppearance(False)
        self.window.progressBar.setRange(0,100)
        self.window.progressBar.reset()
        
    def progressbar_run(self):
        self.window.progressBar.setRange(0,100)
        self.window.progressBar.setFormat("Looking…")
        self.window.progressBar.reset()
        self.timeline.start()
        
    def spinbox_init(self, spin, *, value = 0, min=0, max=1000, step=1, on_change = None):
        spin.setRange(min,max)
        spin.setValue(value)
        if on_change:
            spin.valueChanged.connect(on_change)
        
    def size_changed(self, value):
        self.param.size = value
        self.update_pyramid()
        self.check_size_difficulty()

    def check_size_difficulty(self):
        if self.param.size < 5:
            self.window.difficultyBox.setMaximum(3)
        else:
            self.window.difficultyBox.setMaximum(5)
            
    def difficulty_changed(self, value):
        self.param.difficulty = value
    def min_changed(self, value):
        self.param.value_min = value
        self.window.maxBox.setMinimum(value+1)
    def max_changed(self, value):
        self.param.value_max = value
        self.window.minBox.setMaximum(value-1)

    def update_images(self):
        if not self.images:
            return
        
        if 'puzzle' in self.images:
            self.puzzlePix = toqpixmap(self.images['puzzle'])
        if 'solution' in self.images:
            self.solutionPix = toqpixmap(self.images['solution'])
        if 'nopuzzle' in self.images:
            self.puzzlePix = toqpixmap(self.images['nopuzzle'])
            self.solutionPix = toqpixmap(self.images['nopuzzle'])

        self.window.puzzleImageLabel.setPixmap(self.puzzlePix)
        self.window.solutionImageLabel.setPixmap(self.solutionPix)
        
    def generate_puzzle(self):
        # in another thread. That can be long.
        if not self.generating:
            self.generating = True
            self.progressbar_run()  
            self.information("Looking for a puzzle.")
            self.clean_images()
            self.myThread = LookingForPuzzleThread(self)
            self.myThread.finished.connect(self.when_get_puzzle)
            self.myThread.start(QThread.LowPriority)
            self.puzzleButtonText = self.window.puzzleButton.text()
            self.window.puzzleButton.setText("&Cancel")
        else:
            # cancel the running puzzle generation
            self.generating = False
            self.pyramid.cancel = True

    def when_get_puzzle(self):
        if self.pyramid.cancel:
            self.information("Operation cancelled.")
        else:
            self.redraw_images()

        self.generating = False
        self.window.puzzleButton.setText(self.puzzleButtonText)
        self.progressbar_stop()
        del self.myThread
        
    def information(self, text):
        self.window.informationLabel.setText(str(text))
        
    def clean_images(self):
        self.pyramid.puzzle = {}
        self.pyramid.solution = []
        self.redraw_images()

    def choose_directory(self):
        dirDialog = QFileDialog()
        # default irectory to "Documents"
        dirDialog.setDirectory(self.param.output_directory)
        self.param.output_directory = dirDialog.getExistingDirectory(self.window, "Select Directory")

    def save_puzzle(self):
        combo = self.window.outputTypeCombo
        printerClass = combo.currentData()
        printerClass(self.pyramid, self.param).print()
        
    def validate_pref(self):
        ts_index = self.pref_ui.timestampCombo.currentIndex()
        lw = self.pref_ui.linewidthBox.value()
        resolution = int(self.pref_ui.resolutionCombo.currentText())
        basename = self.pref_ui.basenameEdit.text()
        
        self.param.use_timestamp = True if ts_index == 0 else False
        self.param.line_width = lw
        self.param.resolution = resolution
        if basename:
            self.param.basename = basename
        self.redraw_images()
            
    def redraw_images(self):
        #redraw images with new parameters
        self.imagePrinter.print()
        self.images = self.imagePrinter.get_images()
        self.update_images()    

        
        
        



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat May 30 00:46:19 2020

@author: yves

    This file is part of Saqqarah.

    Saqqarah is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Saqqarah is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Saqqarah.  If not, see <https://www.gnu.org/licenses/>

"""

# Package version
__version__ = "0.9.8b"
__codename__ = "Sanakht (beta version)"

def version():
    return __version__

def codename():
    return __codename__



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun May 31 13:01:37 2020

@author: yves

    This file is part of Saqqarah.

    Saqqarah is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Saqqarah is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Saqqarah.  If not, see <https://www.gnu.org/licenses/>
"""

from . import PyramidPrinter
from . import PrinterCoordinates
from . import fonts
from PIL import Image, ImageDraw, ImageFont
#from pathlib import Path

try:
    import importlib.resources as pkg_resources
except ImportError:
    # Try backported to PY<37 `importlib_resources`.
    import importlib_resources as pkg_resources
    
    
def get_size_from_param(param):
    # param.unit is already set to pixel ?

    size = PrinterCoordinates(param)
    size.xy =  ( param.size * param.boxw + 2*param.outw,
                 param.size * param.boxh + 2*param.outh
                 )

    return round(size)

def center_text(draw, bounding_box, msg, *, font=None, fill='black'):
    x1, y1, x2, y2 = bounding_box

    # Calculate the width and height of the text to be drawn, given font size
    w, h = draw.textsize(msg, font=font)
    
    # Calculate the mid points and offset by the upper left corner of the bounding box
    x = (x2 - x1 - w)/2 + x1
    y = (y2 - y1 - h)/2 + y1

    # Write the text to the image, where (x,y) is the top left corner of the text
    draw.text((x, y), msg, align='center', font=font, fill=fill)
    
class PyramidPrinterImage(PyramidPrinter):
    """
    Implements prining as image
    """
    def __init__(self, pyramid, param, *, memory_images=False):
        super().__init__(pyramid, param)
        Image.init()
        self.param.extensions_list = [ext for ext in Image.SAVE]
        self.param.default_extension = 'png' if 'PNG' in Image.SAVE else list(Image.SAVE.keys())[0].lower()
        self.unit = 'px'
        self.param.outw = 1
        self.param.outh = 1
        
        # introspection
        # needed to load the ttf font file
        with pkg_resources.path(fonts, 'AerialBd.ttf') as font_file:
            # ttf font file. 
            self.param.ttf_file = str(font_file)

        self.param.images = {}
        #unset at end
        self.memory_images = memory_images
        
    def get_images(self):
        return self.param.images
            
    class PrinterCM(PyramidPrinter.PrinterCM):
        def __init__(self, printer):
            self.printer = printer
            self.param = printer.param
            self.param.images = {}
            
        def __enter__(self):
            return self
        
        def __exit__(self, *args):
            pass
        
    class PyramidCM(PyramidPrinter.PyramidCM):
        def __init__(self, printer):
            self.printer = printer
            self.param = printer.param
            self.stream = self.param.stream.lower()
            
        def __enter__(self):
            # Create blank rectangle to write on
            # round return a tuple of rounded values
            im_size = get_size_from_param(self.param)
            self.image = Image.new('RGB', im_size, 'white')
            self.draw = ImageDraw.Draw(self.image)
            return self
        
        def __exit__(self, *args):
            # self.image.show()
            if not self.printer.memory_images:
                filename = self.param.get_filename(self.stream)
                self.image.save(filename)  
                self.param.log(f"Saved image {filename}.")
            else:
                self.param.images[self.stream] = self.image

    def print_pyramid(self, param):
        for i in range(param.size):
            y = i 
            self.__print_line__(y, i+1, param)

    def __print_line__(self, y, no, param):
        for x in range(param.size-no,param.size+no,2):
            self.__print_box__(x,y, param)
            self.__print_node__(x,y, param)
            
    def __print_box__(self, x, y, param):
        draw = param.pyramid_printer.draw
        c1, c3, = (PrinterCoordinates(param) for c in range(2))
        c1.xy = x, y
        c3.xy = x + param.boxw, y + param.boxh
        # width 5 pt
        # TODO: add a parameter
        width = round(self.param.pt(self.param.line_width))
        draw.rectangle([c1.x, c1.y, c3.x, c3.y], outline='black', width=width)

    def __print_node__(self, x, y, param):
        if not param.values:
            return
        value = param.values.pop()
        if value == "_":
            return
        draw = param.pyramid_printer.draw
        
        c1, c2 = (PrinterCoordinates(param) for c in range(2))
        c1.xy = x, y
        c2.xy = x + param.boxw, y + param.boxh
        
        # puzzle
        if type(value) is str:
            color='red'
        else:
            color='black'
        
        fontsize = param.pt(10)
        font = ImageFont.FreeTypeFont(str(self.param.ttf_file), size=fontsize, index=0, encoding="")
        
        center_text(draw, [c1.x, c1.y, c2.x, c2.y], str(value), font=font, fill=color)
    

    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun May 31 21:25:09 2020

@author: yves

    This file is part of Saqqarah.

    Saqqarah is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Saqqarah is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Saqqarah.  If not, see <https://www.gnu.org/licenses/>
"""

from PySide2.QtCore import QStandardPaths as sp
from time import strftime
from pathlib import Path

from . import filename_prepare

class Parameters:
    # output parameter
    basename = 'pyramid'
    __extension = None
    default_extension = None
    extensions_list = []
    
    use_timestamp = True
    
    # Pyramid parameter
    size = 5
    size_min = 4
    size_max = 12
    difficulty = 2
    difficulty_min = 1
    difficulty_max = 5
    seed = None
    value_min = 0
    value_max = 9
    exclude_zero = False
    
    # Printing parameter

    # unit cm
    boxw = 2
    outw = 0
    boxh = 1
    outh = 0
    # pt
    line_width = 1
    
    # ppp
    __resolution = 300
    
    # Tikz
    latex_separator = ''

    units = {
            'cm' : 1,
            'mm':  1/10,
            'inch': 2.54,
            'pt': 2.54/72.27,
            # change it changing resolution
            'px': 2.54/__resolution
            }
    
    scale = 1
    unit = 'cm'
    
    def __init__(self, **kwargs):
        self.__output_directory = sp.writableLocation(sp.DocumentsLocation) or '.'
        for k, v in kwargs:
            setattr(self,k,v)
    
    def u(self, length, *, to_unit = None, scale = None):
        """
        change value cm to unit
        and scale by self.scale
        """
        to_unit = to_unit or self.unit
        scale = scale or self.scale
        return length*scale/self.units[to_unit]
    
    def pt(self,val):
        """
        change from pt to unit
        used for text size and line width
        """
        return round(self.u(val*2.54/72.27))
    
    def n_values(self):
        """ 
        return the numbers of values in the pyramid
        """
        return self.size*(self.size+1)//2

    @property
    def resolution(self):
        return self.__resolution

    @resolution.setter
    def resolution(self, r):
        # ignore if none
        try:
            r = int(r)
        except:
            return
        if r:
            self.__resolution = r
            self.units['px'] = 2.54/r
    @property
    def extension(self):
        return self.__extension if self.__extension else self.default_extension
    
    @extension.setter
    def extension(self, ext):
        self.__extension = ext
    
    @property
    def filename(self):
        pext = f".{self.ext}" if self.ext else ''
        if self.use_timestamp:
            timestamp = strftime('-%Y%m%dT%H%M%S')
            return str(Path(self.output_directory, f"{self.basename}{timestamp}{pext}"))
        
    @filename.setter
    def filename(self, name):
        if name:
            self.basename, self.extension, dirname = filename_prepare(name, self.extensions_list, self.default_extension)
            self.output_directory = dirname if dirname else self.output_directory
        
    def get_filename(self, complement = None):
        ext = self.extension or self.default_extension
        pcomplement = f"-{complement}" if complement else ''
        if self.use_timestamp:
            timestamp = strftime('-%Y%m%dT%H%M%S')
        else:
            timestamp = ''
        return str(Path(self.output_directory, f"{self.basename}{timestamp}{pcomplement}.{ext}"))
    @property
    def output_directory(self):
        return self.__output_directory
    
    @output_directory.setter
    def output_directory(self, dirname):
        if dirname:
            pdir = Path(dirname)
            if pdir.is_dir():
                self.__output_directory = dirname
            
class PrinterCoordinates:
    """
    Automagicly change x et y properties to
    unit given by parameters class
    """
    def __init__(self, parameters):
        self.param = parameters
        self.shift = self.param.outw or self.param.outh

    def __str__(self):
        return f" {self.__x}, {self.__y} "

    def __iter__(self):
        yield self.__x
        yield self.__y
        
    __x = None
    __y = None
    __shift = False
    __shift_x = 0
    __shift_y = 0


    def __round__(self):
        return round(self.__x), round(self.__y)

    @property
    def shift(self):
        return self.__shift
    
    @shift.setter
    def shift(self, bool):
        self.__shift = bool
        if bool:
            self.__shift_x = self.param.u(self.param.outw)
            self.__shift_y = self.param.u(self.param.outh)
        else:
            self.__shift_x = 0
            self.__shift_y = 0
            

    @property
    def xy(self):
        return self.__x+self.__shift_x, self.__y + self.__shift_y
    
    @xy.setter
    def xy(self, coord):
        self.x = coord[0]
        self.y = coord[1]
        
    @property
    def x(self):
        return self.__x+self.__shift_x
    
    @x.setter
    def x(self, val):
        self.__x = self.param.u(val)
        
    @property
    def y(self):
        return self.__y+self.__shift_y

    @y.setter
    def y(self, val):
        self.__y = self.param.u(val)
        



# minimal setup.py to install in develop mode

from setuptools import setup, find_packages

REQUIRED_MODULES = [
    'numpy',
    'Pillow',
    'PySide2'
]

from saqqarah import version

__version__ = version()

# read the contents of the README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name="saqqarah",
    version=__version__,
    long_description=long_description,
    long_description_content_type='text/markdown',   
    author="Yves Combe",
    author_email="yves@ycombe.net",
    url = "https://framagit.org/ycombe/saqqarah",
    description="Calculation pyramid generator",
    packages=find_packages(),
    classifiers=['Development Status :: 3 - Alpha',
                 'Environment :: Console',
                 'Environment :: X11 Applications :: Qt',
                 'Intended Audience :: Education',
                 'License :: DFSG approved',
                 'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
                 'Topic :: Education',
                 ],
    package_data={'': ['fonts/*.ttf', 'ui/*.ui', '*.txt', 'LICENSE']},
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'saqqarah = saqqarah.__main__:main',
        ],
        'gui_scripts' : [
            'saqqarahW = saqqarah.__QT5__:main'
        ]
    },
    install_requires=REQUIRED_MODULES,
    python_requires='>=3.6',
    keywords=["mathematics", "addition", "pyramid","school"]
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun  3 17:59:23 2020

@author: yves


    This file is part of Saqqarah.

    Saqqarah is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Saqqarah is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Saqqarah.  If not, see <https://www.gnu.org/licenses/>
"""

import sys
from PySide2 import QtCore, QtGui, QtWidgets
from PySide2.QtUiTools import QUiLoader
from . import ui

import saqqarah

try:
    import importlib.resources as pkg_resources
except ImportError:
    # Try backported to PY<37 `importlib_resources`.
    import importlib_resources as pkg_resources

def mainwindow_setup(w):
    w.setWindowTitle("Saqqarah ")

def print_dict(object):
    for k,v in object.__dict__.items():
        print(f"{k} :{v}")
        
def main():
    loader = QUiLoader()
    app = QtWidgets.QApplication.instance()
    if app is None: 
        app = QtWidgets.QApplication(sys.argv)

    with pkg_resources.path(ui, 'saqqarah-qt.ui') as ui_file:
        window = loader.load(str(ui_file), None)
        uicontrol = saqqarah.SaqqarahUI(window) 
        mainwindow_setup(window)
        window.show()

    #sys.exit(app.exec_())
    sys.exit(app.exec_())




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri May 29 23:48:15 2020

@author: yves


    This file is part of Saqqarah.

    Saqqarah is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Saqqarah is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Saqqarah.  If not, see <https://www.gnu.org/licenses/>
"""

import argparse

from saqqarah import Pyramid,PyramidPrinterTikz, PyramidPrinterImage, Parameters, PrinterCoordinates, version, codename

def main():
    parser = argparse.ArgumentParser(description=f"Calculation pyramid generator, version {version()} {codename()}" )
    parser.add_argument("-s", "--size", type=int, default=5,
                    help="Piramid size (4 or more) (default: 5)")
    parser.add_argument("-d", "--difficulty", type=int, choices=range(1,6), default=2,
                    help="difficlty level (max 3 if size is 4) (default: 2)")
    parser.add_argument("-m", "--min", type=int, default=0,
                    help="minimum number in base (default: 0)")    
    parser.add_argument("-M", "--max", type=int, default=9,
                    help="maximum number in base (default: 9)")    
    parser.add_argument("-z", "--non-zero", dest='exclude_zero', default=False, action='store_true',
                    help="exclude zeros from base numbers (default: False)")
    parser.add_argument("-S", "--seed", type=int,
                    help="Seed number for random generator")
    parser.add_argument("-E", "--empty", dest='empty', default=False, action='store_true',
                    help="Print an empty pyramid")
    parser.add_argument('--version', action='version', version=f'%(prog)s {version()} {codename()}')
    subparsers = parser.add_subparsers(help='choose output mode (without: list the boxes & values)')
    parser_img = subparsers.add_parser('image', help=f'generate image output ("%(prog)s image -h" for help)')
    parser_img.set_defaults(Printer=PyramidPrinterImage)
    parser_img.add_argument("-f", "--filename", type= str, 
                    help="output schema (default: pyramid-TIME.png gives pyramid-TIME-puzzle.png and pyramid-TIME-solution.png)")
    parser_img.add_argument("-r", "--resolution", type=str,
                    help="resolution of the final images.")
    parser_img.add_argument("-d", "--directory", dest="output_directory",  type=str,
                            help="directory where write images")
    parser_img.add_argument("-t", "--no-timestamp", dest='use_timestamp', default=True, action='store_false',
                    help="Don't add timestamp in file name")
    parser_tikz = subparsers.add_parser('tikz', help=f'generate tikz output ("%(prog)s image -h" for help)')
    parser_tikz.set_defaults(Printer=PyramidPrinterTikz)
    parser_tikz.add_argument("-f", "--filename", type= str,
                    help="output name (default: pyramid-TIMESTAMP.tex)")
    parser_tikz.add_argument("-l", "--latex-separator", type=str, default="",
                    help="latex code to insert between puzzle and solution")
    parser_tikz.add_argument("-d", "--directory", dest="output_directory", type=str,
                            help="directory where write images")
    parser_tikz.add_argument("-t", "--no-timestamp", dest='use_timestamp', default=True, action='store_false',
                    help="Don't add timestamp in file name")
    
    args = parser.parse_args()

    # print debugging
    #print(args.__dict__)

    pyra = Pyramid(args.size)
    if not args.empty:
        pyra.get_puzzle(args.difficulty,args.min, args.max, seed=args.seed, exclude_zero=args.exclude_zero)

    # defaut directory is "Documents"
    # but in command line, default should be '.'
    if hasattr(args, 'args.output_directory'):
        args.output_directory = args.output_directory or '.'
    
    if hasattr(args, 'Printer'):
        param = Parameters()
        # initialise param before pass arguments
        pyraPr = args.Printer(pyra, param)

        for k,v in args.__dict__.items():
            setattr(param, k, v)

        pyraPr.print()
    
    else:
        print("""
        Boxes numeroted from top left to bottom.
        Going from left to right in each line.\n
        """)
        
        print(f"Puzzle for pyramid size {args.size} :")
        for box, value in pyra.puzzle.items():
            print(f"     box : {box+1:2d}   value : {value:4d} ")
        print(f"\nSolution :")
        for box, value in enumerate(pyra.solution):
            print(f"     box : {box+1:2d}   value : {int(value):4d} ")
        
    try:
        exit(1)
    except NameError:
        pass


